from abc import ABC, abstractmethod
from typing import Optional
import numpy as np


from .ClusterScorer import ClusterScorer


class OutlierDetector(ABC):
    def __init__(self, outlier_scorer: ClusterScorer):
        self.outlier_scorer = outlier_scorer

    @abstractmethod
    def detect_outliers(
        self, embeddings: np.ndarray, labels: list[int]
    ) -> tuple[Optional[np.ndarray], dict[int, list[float]]]:
        pass



from sklearn.neighbors import LocalOutlierFactor
from sklearn.metrics import silhouette_score
from typing import Optional
import numpy as np


from ..abstract.OutlierDetector import OutlierDetector
from ..abstract.ClusterScorer import ClusterScorer


class LOFOutlierDetector(OutlierDetector):
    def __init__(
        self,
        outlier_scorer: ClusterScorer,
        n_neighbors_range: tuple[int, int],
        step: int,
    ):
        super().__init__(outlier_scorer)
        self.n_neighbors_range = n_neighbors_range
        self.step = step

    def detect_outliers(
        self, embeddings: np.ndarray, labels: list[int]
    ) -> tuple[Optional[np.ndarray], dict[int, list[float]]]:
        # FIXME: remove unused variable
        # best_n_neighbors = None
        best_silhouette_score = -1
        best_outlier_mask = None
        scores = self.outlier_scorer.calculate_scores(embeddings, labels)

        for n_neighbors in range(
            self.n_neighbors_range[0], self.n_neighbors_range[1] + 1, self.step
        ):
            lof = LocalOutlierFactor(n_neighbors=n_neighbors, contamination="auto")
            outlier_scores = lof.fit_predict(embeddings)
            outlier_mask = outlier_scores == -1

            silhouette_avg = silhouette_score(
                embeddings[~outlier_mask], labels[~outlier_mask]
            )
            if silhouette_avg > best_silhouette_score:
                best_silhouette_score = silhouette_avg
                # FIXME: remove unused variable
                # best_n_neighbors = n_neighbors
                best_outlier_mask = outlier_mask

        return best_outlier_mask, scores



import random
from sklearn.metrics import adjusted_rand_score
from itertools import combinations
from sklearn.metrics import silhouette_score
from sklearn.metrics import normalized_mutual_info_score



def get_cluster_pairs(uuids, labels):
    pairs = set()
    for cluster in set(labels):
        if cluster != -1:  # ignore unclustered items (noise)
            cluster_uuids = [uuid for uuid, label in zip(uuids, labels) if label == cluster]
            pairs.update(combinations(cluster_uuids, 2))
    return pairs

def calculate_rates(predicted_pairs, expected_pairs, total_pairs):
    false_positives = len(predicted_pairs - expected_pairs)
    false_negatives = len(expected_pairs - predicted_pairs)
    true_positives = len(predicted_pairs & expected_pairs)
    true_negatives = len(total_pairs - predicted_pairs - expected_pairs)
    
    false_positive_rate = false_positives / (false_positives + true_negatives) if (false_positives + true_negatives) > 0 else 0
    false_negative_rate = false_negatives / (false_negatives + true_positives) if (false_negatives + true_positives) > 0 else 0
    
    return false_positive_rate, false_negative_rate, true_positives, false_positives, false_negatives

def calculate_metrics(true_positives, false_positives, false_negatives):
    precision = true_positives / (true_positives + false_positives) if (true_positives + false_positives) > 0 else 0
    recall = true_positives / (true_positives + false_negatives) if (true_positives + false_negatives) > 0 else 0
    f1_score = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0
    
    return precision, recall, f1_score


def evaluate(uuids, predicted_labels, expected_labels, embeddings=None):
    predicted_pairs = get_cluster_pairs(uuids, predicted_labels)
    expected_pairs = get_cluster_pairs(uuids, expected_labels)
    total_pairs = set(combinations(uuids, 2))
    
    false_positive_rate, false_negative_rate, true_positives, false_positives, false_negatives = calculate_rates(
        predicted_pairs, expected_pairs, total_pairs
    )
    
    precision, recall, f1_score = calculate_metrics(true_positives, false_positives, false_negatives)
    
    ari = adjusted_rand_score(expected_labels, predicted_labels)
    nmi = normalized_mutual_info_score(expected_labels, predicted_labels)
    
    if embeddings is not None and len(set(predicted_labels)) > 1:
        sil_score = silhouette_score(embeddings, predicted_labels)
    else:
        sil_score = float('nan')
    
    valid_scores = [metric for metric in [f1_score, ari, sil_score, nmi, (1-false_negative_rate)] if not (isinstance(metric, float) and metric != metric)]
    final_score = sum(valid_scores) / len(valid_scores)

    return {
        "false_positive_rate": false_positive_rate,
        "false_negative_rate": false_negative_rate,
        "precision": precision,
        "recall": recall,
        "f1_score": f1_score,
        "silhouette_score": sil_score,
        "adjusted_rand_index": ari,
        "normalized_mutual_information": nmi,
        "final_score": final_score
    }


import uuid
def create_dummy_data():
    uuids = [str(uuid.uuid4()) for _ in range(10)]
    
    # predicted labels: 3 clusters and 1 outlier
    predicted_labels = [0, 0, 0, 1, 1, 2, 2, 2, -1, 1]
    
    # expected labels: 2 clusters and 2 outliers
    expected_labels = [3, 3, 3, 2, 2, 2, 2, -1, -1, 3]

    expected_labels = predicted_labels

    expected_labels = [random.randint(0, 5) for _ in range(10)]
    
    return uuids, predicted_labels, expected_labels





from sklearn.metrics import silhouette_samples
import numpy as np

from ..abstract.ClusterScorer import ClusterScorer


class SilhouetteClusterScorer(ClusterScorer):
    def calculate_scores(
        self, embeddings: np.ndarray, labels: np.ndarray
    ) -> dict[int, list[float]]:
        silhouette_scores = silhouette_samples(embeddings, labels)
        scores = {}
        for label in np.unique(labels):
            cluster_mask = labels == label
            scores[label] = silhouette_scores[cluster_mask].tolist()
        return scores



from abc import ABC, abstractmethod
import numpy as np


class ClusterScorer(ABC):
    @abstractmethod
    def calculate_scores(
        self, embeddings: np.ndarray, labels: list[int]
    ) -> dict[int, list[float]]:
        pass



from sklearn.cluster import KMeans
import numpy as np

from ..abstract.ClusteringAlgorithm import ClusteringAlgorithm


class KMeansClustering(ClusteringAlgorithm):
    def __init__(self, config: dict):
        self.config = config

    def cluster(self, embeddings: np.ndarray) -> list[int]:
        kmeans = KMeans(**self.config)
        labels = kmeans.fit_predict(embeddings)

        return labels



# from sklearn.manifold import TSNE
# # from umap import UMAP
# import pandas as pd
# import matplotlib.pyplot as plt
# import numpy as np
# from typing import Optional


# from ..abstract.ClusterScorer import ClusterScorer


# class ResultVisualizer:
#     def __init__(
#         self,
#         config: dict,
#         outlier_scorer: ClusterScorer,
#         cluster_scores: Optional[list[int]] = None,
#     ):
#         self.config = config
#         self.outlier_scorer = outlier_scorer
#         self.labels: Optional[list[int]] = None
#         self.embeddings: Optional[np.ndarray] = None
#         self.data: Optional[list[str]] = None
#         self.cluster_scores = cluster_scores
#         self.outlier_scores: Optional[dict[int, list[float]]] = None

#     def visualize(
#         self,
#         clusters: list[int],
#         embeddings: np.ndarray,
#         data: list[str],
#         method="tsne",
#     ):
#         self.labels = clusters
#         self.embeddings = embeddings
#         self.data = data

#         reduced_embeddings = self._reduce_dimensions(embeddings, method)
#         self._plot_clusters(reduced_embeddings, clusters, method)

#     def _reduce_dimensions(self, embeddings: np.ndarray, method: str) -> np.ndarray:
#         # if method == "tsne":
#         #     reducer = TSNE(n_components=2, random_state=42)
#         # elif method == "umap":
#         #     reducer = UMAP(n_components=2, random_state=42)
#         # else:
#         #     raise ValueError(f"Invalid visualization method: {method}")

#         # reduced_embeddings = reducer.fit_transform(embeddings)
#         # return reduced_embeddings

#     def _plot_clusters(
#         self, reduced_embeddings: np.ndarray, clusters: list[int], method: str
#     ):
#         fig, ax = plt.subplots(figsize=(8, 6))
#         scatter = ax.scatter(
#             reduced_embeddings[:, 0],
#             reduced_embeddings[:, 1],
#             c=clusters,
#             cmap="viridis",
#         )
#         legend = ax.legend(*scatter.legend_elements(), title="Clusters")
#         ax.add_artist(legend)
#         ax.set_title(f"{method.upper()} Visualization")
#         plt.show()

#     def print_clusters(
#         self,
#         clusters: list[int],
#         embeddings: np.ndarray,
#         data: pd.DataFrame,
#         title_col: str,
#         content_col: str,
#         sort_by="mean_similarity",
#     ):
#         self.labels = clusters
#         self.embeddings = embeddings
#         self.data = data

#         cluster_metrics = self._calculate_cluster_metrics()
#         sorted_clusters = self._sort_clusters(cluster_metrics, sort_by)

#         self.outlier_scores = self.outlier_scorer.calculate_scores(
#             self.embeddings, self.labels
#         )

#         assert self.data is not None

#         for label, metrics in sorted_clusters:
#             cluster_mask = self.labels == label
#             cluster_data = self.data[cluster_mask]
#             cluster_embeddings = self.embeddings[cluster_mask]
#             cluster_outlier_scores = self.outlier_scores[label]

#             self._print_cluster_details(
#                 label,
#                 metrics,
#                 cluster_data,
#                 cluster_embeddings,
#                 title_col,
#                 content_col,
#                 cluster_outlier_scores,
#             )

#     def _print_cluster_details(
#         self,
#         label: int,
#         metrics: dict[str, float],
#         cluster_data: pd.DataFrame,
#         cluster_embeddings: np.ndarray,
#         title_col: str,
#         content_col: str,
#         cluster_outlier_scores: list[float],
#     ):
#         print(
#             f"Cluster {label} (Mean Similarity: {metrics['mean_similarity']:.2f}, "
#             f"Silhouette Score: {metrics['silhouette_score']:.2f}):"
#         )

#         sorted_indices = np.argsort(cluster_outlier_scores)[::-1]

#         for i in sorted_indices:
#             row = cluster_data.iloc[i]
#             combined_text = f"{row[title_col]}".replace("\n", " ").strip()
#             outlier_score = cluster_outlier_scores[i]
#             is_outlier = self.outlier_mask[np.where(self.labels == label)[0][i]]
#             if is_outlier:
#                 print(f"Score.: {outlier_score:.2f}) [OUTLIER]  {combined_text}")
#             else:
#                 print(f"Score.: {outlier_score:.2f})            {combined_text}")

#         print()

#     def _calculate_cluster_metrics(self) -> dict[int, dict[str, float]]:
#         assert self.labels is not None
#         assert self.outlier_scores is not None
#         assert self.cluster_scores is not None

#         cluster_metrics = {}
#         for label in np.unique(self.labels):
#             avg_scores = self.outlier_scores[label]
#             # print('-----------------')
#             # print(avg_scores)
#             # print('-----------------')

#             mean_similarity = np.mean(avg_scores)
#             silhouette_score = self.cluster_scores[label]
#             cluster_metrics[label] = {
#                 "mean_similarity": mean_similarity,
#                 "silhouette_score": silhouette_score,
#             }
#         return cluster_metrics

#     def _sort_clusters(
#         self, cluster_metrics: dict[int, dict[str, float]], sort_by: str
#     ) -> list[tuple[int, dict[str, float]]]:
#         if sort_by == "mean_similarity":
#             return sorted(
#                 cluster_metrics.items(),
#                 key=lambda x: x[1]["mean_similarity"],
#                 reverse=True,
#             )
#         elif sort_by == "silhouette_score":
#             return sorted(
#                 cluster_metrics.items(),
#                 key=lambda x: x[1]["silhouette_score"],
#                 reverse=True,
#             )
#         elif sort_by == "combined":
#             return sorted(
#                 cluster_metrics.items(),
#                 key=lambda x: (x[1]["mean_similarity"] + x[1]["silhouette_score"]) / 2,
#                 reverse=True,
#             )
#         else:
#             raise ValueError(f"Invalid sorting criterion: {sort_by}")

#     def set_outlier_mask(self, outlier_mask: np.ndarray):
#         self.outlier_mask = outlier_mask



# from sklearn.manifold import TSNE
# # from umap import UMAP
# import numpy as np
# import matplotlib.pyplot as plt


# from .ResultVisualizer import ResultVisualizer


# class MultiResultVisualizer(ResultVisualizer):
#     def __init__(self, config: dict, cluster_scores: list[int] = None):
#         super().__init__(config, cluster_scores)

#     def visualize(
#         self,
#         clusters: list[int],
#         embeddings: np.ndarray,
#         data: list[str],
#         method="tsne",
#     ):
#         self.labels = clusters
#         self.embeddings = embeddings
#         self.data = data

#         # if method == "tsne":
#         #     tsne = TSNE(n_components=2, random_state=42)
#         #     reduced_embeddings = tsne.fit_transform(embeddings)
#         # elif method == "umap":
#         #     umap_reducer = UMAP(n_components=2, random_state=42)
#         #     reduced_embeddings = umap_reducer.fit_transform(embeddings)
#         # else:
#         #     raise ValueError(f"Invalid visualization method: {method}")

#         fig, ax = plt.subplots(figsize=(8, 6))
#         scatter = ax.scatter(
#             reduced_embeddings[:, 0],
#             reduced_embeddings[:, 1],
#             c=clusters,
#             cmap="viridis",
#         )
#         legend = ax.legend(*scatter.legend_elements(), title="Clusters")
#         ax.add_artist(legend)
#         ax.set_title(f"{method.upper()} Visualization")
#         plt.show()

#     def print_clusters(
#         self, clusters: list[int], embeddings: np.ndarray, data: List[str]
#     ):
#         self.labels = clusters
#         self.embeddings = embeddings
#         self.data = data

#         for label, score in sorted(
#             self.cluster_scores.items(), key=lambda x: x[1], reverse=True
#         ):
#             print(f"Cluster {label} (Silhouette Score: {score}):")
#             self._print_cluster(label)
#             print()

#     def _print_cluster(self, label):
#         cluster_data = [
#             self.data[i] for i in range(len(self.data)) if self.labels[i] == label
#         ]
#         for doc in cluster_data:
#             print(doc[0].replace("\n", "").strip())



import numpy as np


class EmbeddingTimestamper:
    def __init__(self, config: dict):
        self.config = config

    def timestamp(self, embeddings: np.ndarray, timestamps: list[float]) -> np.ndarray:
        # Normalize timestamps to be between 0 and 1
        min_timestamp = min(timestamps)
        max_timestamp = max(timestamps)
        normalized_timestamps = [
            (t - min_timestamp) / (max_timestamp - min_timestamp) for t in timestamps
        ]

        # Add timestamp dimension to the embeddings
        timestamped_embeddings = np.hstack(
            (embeddings, np.array(normalized_timestamps)[:, np.newaxis])
        )

        return timestamped_embeddings



import numpy as np
from typing import Optional


from ..abstract.OutlierDetector import OutlierDetector
from ..abstract.ClusterScorer import ClusterScorer


class ZScoreOutlierDetector(OutlierDetector):
    def __init__(
        self, outlier_scorer: ClusterScorer, threshold: float, min_score: float
    ):
        super().__init__(outlier_scorer)
        self.threshold = threshold
        self.min_score = min_score

    def detect_outliers(
        self, embeddings: np.ndarray, labels: list[int]
    ) -> tuple[Optional[np.ndarray], dict[int, list[float]]]:
        scores = self.outlier_scorer.calculate_scores(embeddings, labels)
        outlier_mask = np.full(len(labels), False)
        for label, cluster_scores in scores.items():
            mean_score = np.mean(cluster_scores)
            std_score = np.std(cluster_scores)
            cluster_mask = labels == label

            z_scores = np.divide((np.array(cluster_scores) - mean_score), std_score, out=np.zeros_like(cluster_scores), where=std_score!=0)
            outlier_mask[cluster_mask] = (z_scores < -self.threshold) & (
                np.array(cluster_scores) < self.min_score
            )

        return outlier_mask, scores



import numpy as np
import pandas as pd
# from umap import UMAP
from typing import Optional


from .abstract.ClusterScorer import ClusterScorer
from .abstract.OutlierDetector import OutlierDetector
from .embedding_generators.SentenceTransformerEmbeddingGenerator import (
    SentenceTransformerEmbeddingGenerator,
)
from .cluster_scorers.SilhouetteClusterScorer import SilhouetteClusterScorer
from .cluster_scorers.CosineSimililarityClusterScorer import (
    CosineSimililarityClusterScorer,
)
from .clustering_algorithms.HDBSCANClustering import HDBSCANClustering
# from .result_visualizers.ResultVisualizer import ResultVisualizer
from .misc.EmbeddingTimestamper import EmbeddingTimestamper


class ClusteringPipeline:
    def __init__(
        self,
        config: dict,
        model_name: str,
        uuid_col: str,
        datetime_col: str,
        outlier_scorer: ClusterScorer,
        outlier_detector: OutlierDetector,
        # FIXME: unused variable
        embedding_col: Optional[str] = None,

        title_col: Optional[str] = None,
        description_col: Optional[str] = None,
        content_col: Optional[str] = None,

        verbose: bool = False,
    ):
        # if embedding_col is None and title_col is None:
        #     raise ValueError("Either 'embedding_col' or 'title_col' must be provided")

        self.config = config

        self.embedding_generator = SentenceTransformerEmbeddingGenerator(
            config, model_name
        )
        
        self.clustering_algorithm = HDBSCANClustering(config)
        
        self.dimensionality_reducer = None
        # self.dimensionality_reducer = UMAP(
        #     n_components=100, random_state=42, metric="euclidean"
        # )
        # self.dimensionality_reducer = VAEDimensionalityReducer(input_dim=1025, latent_dim=100)
        # self.dimensionality_reducer = ParametricUMAP(n_components=100, random_state=42)

        # self.visualizer = ResultVisualizer(config, outlier_scorer=outlier_scorer)
        self.embedding_timestamper = EmbeddingTimestamper(config)
        self.uuid_col = uuid_col

        self.title_col = title_col
        self.description_col = description_col
        self.content_col = content_col

        self.datetime_col = datetime_col

        self.raw_embeddings = None
        self.processed_embeddings = None
        self.data = None
        self.labels = None
        self.cluster_scores = None
        self.outlier_scorer = outlier_scorer
        self.outlier_detector = outlier_detector
        self.outlier_mask: Optional[np.ndarray] = None
        self.verbose = verbose

        if (self.verbose):
            print("Pipeline initialized")

    def run(self, data: pd.DataFrame, embeddings=None, process_embeddings=True):
        self._process_data(data, embeddings, process_embeddings)
        return self._run()

    def _process_data(self, data, embeddings, process_embeddings, data_is_list=False):
        self.data = data

        if data_is_list:
            titles = data
        else:
            titles = data[self.title_col].tolist()

        if embeddings is None:
            if (self.verbose):
                print("Generating embeddings")
            embeddings = self._generate_embeddings(titles)
            if (self.verbose):
                print("Embeddings generated")

        self.raw_embeddings = embeddings
        
        # FIXME: add timestamps
        # timestamps = self.get_timestamps(data)

        if process_embeddings:
            self.processed_embeddings = self._process_embeddings(embeddings)
        else:
            self.processed_embeddings = embeddings

    def get_timestamps(self, data):
        datetimes = data[self.datetime_col].tolist()
        timestamps = [dt.timestamp() for dt in datetimes]
        return timestamps

    def run_embeddings_only(self, embeddings_df):
        # embeddings = embeddings_df["title_embed"].values
        self.data = embeddings_df
        embeddings = np.stack(embeddings_df["title_embed"].values)
        # FIXME: add timestamps
        # timestamps = self.get_timestamps(embeddings_df)
        self.processed_embeddings = self._process_embeddings(embeddings)
        return self._run()

    def _run(self):
        # get labels
        self.labels = self._cluster(self.processed_embeddings)

        # detect outliers
        self.detect_outliers(self.raw_embeddings, self.labels)

        # mark outliers as -1
        for i, l in enumerate(self.labels):
            if self.outlier_mask[i]:
                self.labels[i] = -1

        # score clusters
        self.cluster_scores = self._calculate_cluster_scores(self.labels)

        # sort clusters by score
        self.cluster_scores = dict(
            sorted(self.cluster_scores.items(), key=lambda x: x[1], reverse=True)
        )

        return self.labels

    def detect_outliers(self, embeddings: np.ndarray, labels: list[int]):
        if self.outlier_detector is not None:
            outlier_mask, item_scores = self.outlier_detector.detect_outliers(
                embeddings, labels
            )

            if outlier_mask is not None:
                self.outlier_mask = outlier_mask
                self.item_scores = item_scores

        else:
            self.item_scores = self.outlier_scorer.calculate_scores(embeddings, labels)
            self.outlier_mask = np.full(len(labels), False)

        if (self.verbose):
            print("Outliers:", self.outlier_mask)

        return self.outlier_mask, self.item_scores

    def _generate_embeddings(self, data):
        embeddings = self.embedding_generator.generate_embeddings(data)
        return embeddings

    def _process_embeddings(self, embeddings):
        self.raw_embeddings = embeddings
        # timestamped_embeddings = self.embedding_timestamper.timestamp(embeddings, timestamps)
        # timestamped_embeddings = embeddings
        if self.dimensionality_reducer is not None:
            return self.dimensionality_reducer.fit_transform(
                embeddings
            )
        else:
            return embeddings

    def _cluster(self, embeddings):
        clusters = self.clustering_algorithm.cluster(embeddings)
        clusters_array = np.array(clusters)

        if (self.verbose):
            print("Shape of clusters_array:", clusters_array.shape)

        return clusters_array

    def _calculate_cluster_scores(self, labels):
        silhouette = SilhouetteClusterScorer()
        scored_silhouette_samples = silhouette.calculate_scores(
            self.raw_embeddings, labels
        )
        cosine = CosineSimililarityClusterScorer()
        scored_cosine_samples = cosine.calculate_scores(self.raw_embeddings, labels)

        if (self.verbose):
            print(scored_cosine_samples)

        cluster_scores = {}
        unique_labels = np.unique(labels)
        for label in unique_labels:
            # silhouette_scores = scored_silhouestte_samples[label]
            cosine_scores = scored_cosine_samples[label]
            cluster_scores[label] = np.mean(cosine_scores)

        if (self.verbose):
            print(scored_cosine_samples)

        return cluster_scores

    def get_clusters(self, threshold: bool | float = 0.70, include_outliers: bool = False):
        labels = self.labels
        item_dicts = []
        scorer = CosineSimililarityClusterScorer()
        scored_samples = scorer.calculate_scores(self.raw_embeddings, labels)
        if (self.verbose):
            print("Length of labels:", len(labels))
            print("Scored Samples:", scored_samples)

        for i, label in enumerate(labels):
            if label == -1:
                score = -1
            else:
                score = scored_samples[label][0]
                scored_samples[label].pop(0)

            if threshold and score < threshold:
                label = -1

            item_dicts.append(
                {
                    "uuid": self.data.iloc[i][self.uuid_col],
                    "label": int(label),
                    "score": float(score),
                }
            )

        clusters = {}
        for item in item_dicts:
            label = item["label"]

            if label not in clusters:
                # clusters[label] = {"score": item["score"], "items": [item["uuid"]]} OLD
                clusters[label] = {"score": item["score"], "items": [item]}
            else:
                # clusters[label]["items"].append(item["uuid"])        
                clusters[label]["items"].append(item)    

        result = map(lambda x: {**{"label": x}, **clusters[x]}, clusters)
        result = sorted(result, key=lambda x: x["score"], reverse=True)
        result = filter(lambda x: len(x["items"]) > 1, result)
        if not include_outliers:
            result = filter(lambda x: x["label"] != -1, result)

        result = list(result)

        return result

    # def visualize(self, method="tsne"):
    #     self.visualizer.visualize(
    #         self.labels, self.processed_embeddings, self.data, method
    #     )

    def hyperparameter_tuning(self, data: list[str], uuids: list[str], expected_labels: list[int], n_iter_search: int = 20):
        if self.raw_embeddings is None:
            self._process_data(data, None, False, data_is_list=True)

        print("shape of processed embeddings:", self.raw_embeddings.shape)
                
        best_params = self.clustering_algorithm.hyperparameter_tuning(
            self.raw_embeddings, uuids, expected_labels, self.outlier_detector, n_iter_search
        )

        return best_params


# import numpy as np
# from umap import UMAP

# from .result_visualizers.MultiResultVisualizer import MultiResultVisualizer
# from .ClusteringPipeline import ClusteringPipeline


# class MultiSentClusteringPipeline(ClusteringPipeline):
#     """
#     A class representing a double clustering pipeline.
#     Embeds first and second sentences separately and concatenates the embeddings to form the final embedding.
#     Then performs dimensionality reduction and clustering on the final embeddings.

#     Args:
#         config (Dict): A dictionary containing the configuration parameters for the pipeline.
#         model_name (str, optional): The name of the model to be used. Defaults to None.
#     """

#     def __init__(self, config: dict, model_name: str = None):
#         super().__init__(config, model_name)
#         self.dimensionality_reducer = UMAP(n_components=100, random_state=42)
#         self.visualizer = MultiResultVisualizer(config)
#         print("DoubleClusteringPipeline initialized")

#     def run(self, data: list[list[str]]):
#         # self.data = TextPreprocessor(self.config).preprocess(data)
#         self.data = data
#         self.embeddings = self._generate_double_embeddings(self.data)
#         print("Shape of embeddings:", self.embeddings.shape)
#         print("timestamp", self.embeddings[0][-1])
#         return self._run(self.data)

#     def _generate_double_embeddings(self, data: list[list[str]]):
#         print(data[0])
#         sentence_embeddings = []
#         for i in range(len(data[0])):
#             print("Processing sentence", i)
#             sentences = [d[i] for d in data]
#             embeddings = self.embedding_generator.generate_embeddings(sentences)
#             sentence_embeddings.append(embeddings)

#         concatenated_embeddings = np.concatenate(sentence_embeddings, axis=1)
#         if self.dimensionality_reducer is not None:
#             concatenated_embeddings = self.dimensionality_reducer.fit_transform(
#                 concatenated_embeddings
#             )
#         print("Shape of concatenated embeddings:", concatenated_embeddings.shape)

#         return concatenated_embeddings



import time
from typing import Callable, Optional
import numpy as np
from hdbscan import HDBSCAN, all_points_membership_vectors
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.model_selection import ParameterSampler
from ..misc.BenchmarkUtils import evaluate
from ..abstract.ClusteringAlgorithm import ClusteringAlgorithm
from ..cluster_scorers.CosineSimililarityClusterScorer import CosineSimililarityClusterScorer
from sklearn.decomposition import PCA
import umap
from tqdm import tqdm

class HDBSCANClustering(ClusteringAlgorithm):
    def __init__(self, config: dict):
        self.config = config

    def cluster(self, embeddings: np.ndarray) -> list[int]:
        embeddings, config = self._prepare_embeddings(embeddings)
        clusterer = HDBSCAN(**config)
        clusterer.fit(embeddings)
        labels = self._assign_labels(clusterer)
        return labels

    def hyperparameter_tuning(
        self,
        embeddings: np.ndarray,
        uuids: list[str],
        expected_labels: list[int],
        outlier_detector: Optional[Callable] = None,
        n_iter_search: int = 20,
    ) -> dict:
        print("Tuning hyperparameters for HDBSCAN, outlier detection, and dimensionality reduction...")
        param_list = self._generate_parameter_list(n_iter_search)
        best_params_list = self._evaluate_parameters(embeddings, uuids, expected_labels, outlier_detector, param_list)
        self._print_final_results(best_params_list)
        return best_params_list

    def _prepare_embeddings(self, embeddings: np.ndarray) -> tuple:
        config = self.config.copy()
        if self.config["metric"] == "cosine":
            embeddings = cosine_similarity(embeddings)
            config["metric"] = "precomputed"
        config["prediction_data"] = True
        embeddings = embeddings.astype(np.float64)
        return embeddings, config

    def _assign_labels(self, clusterer) -> list[int]:
        soft_clusters = all_points_membership_vectors(clusterer)
        labels_from_probabilities = np.argmax(soft_clusters, axis=1)
        labels = [
            label if label != -1 else labels_from_probabilities[idx]
            for idx, label in enumerate(clusterer.labels_)
        ]
        return labels

    def _generate_parameter_list(self, n_iter_search: int) -> list:
        hdbscan_param_dist, outlier_param_dist, pca_param_dist, _, no_dim_reducer_param_dist = self._define_param_distributions()
        # pca_param_list = list(ParameterSampler(
        #     {**hdbscan_param_dist, **outlier_param_dist, **pca_param_dist},
        #     n_iter=n_iter_search // 2
        # ))
        # umap_param_list = list(ParameterSampler(
        #     {**hdbscan_param_dist, **outlier_param_dist, **umap_param_dist},
        #     n_iter=n_iter_search // 3
        # ))
        no_dim_reducer_param_list = list(ParameterSampler(
            {**hdbscan_param_dist, **outlier_param_dist, **no_dim_reducer_param_dist},
            n_iter=n_iter_search
        ))
        # return pca_param_list + umap_param_list + no_dim_reducer_param_list
        return no_dim_reducer_param_list

    def _define_param_distributions(self) -> tuple:
        hdbscan_param_dist = {
            "min_samples": [5, 6, 7, 8, 9, 10],
            "min_cluster_size": [2],
            "cluster_selection_method": ["eom"],
            "metric": ["euclidean"],
            "algorithm": ["best", "prims_kdtree"],
            "cluster_selection_epsilon": [0.1, 0.15],
            "alpha": [0.5, 0.55, 0.6, 0.65],
            "prediction_data": [True]
        }

        outlier_param_dist = {
            "threshold": [0.6, 0.7, 0.75, 0.8, 0.85, 0.9, 1.0],
            "min_score": [0.8, 0.75, 0.7, 0.65, 0.6]
        }

        pca_param_dist = {
            'reducer': ['pca'],
            'pca__n_components': [25, 50, 100, 150, 300, 500],
            'pca__svd_solver': ['auto', 'full', 'arpack', 'randomized'],
            'pca__whiten': [True, False]
        }
        umap_param_dist = {
            'reducer': ['umap'],
            'umap__n_neighbors': [5, 10, 15],
            'umap__min_dist': [0.1, 0.5, 0.9],
            'umap__n_components': [50, 75, 100, 125, 150]
        }
        no_dim_reducer_param_dist = {
            'reducer': ['none']
        }
        return hdbscan_param_dist, outlier_param_dist, pca_param_dist, umap_param_dist, no_dim_reducer_param_dist

    def _evaluate_parameters(self, embeddings, uuids, expected_labels, outlier_detector, param_list) -> list:
        best_params_list = []
        start_time = time.time()

        for i, params in tqdm(enumerate(param_list), total=len(param_list)):
            # Rest of the code...
            # print(f"Iteration {i+1}/{len(param_list)}")
            # print(f"Trying parameters: {params}")
            try:
                metrics, predicted_labels = self._attempt_clustering(embeddings, params, uuids, expected_labels, outlier_detector)
                # score = aggregate_metrics(metrics)
                score = metrics['final_score']
                best_params_list.append({'params': params, 'score': score, 'metrics': metrics})
                best_params_list.sort(key=lambda x: x['score'], reverse=True)
                if len(best_params_list) > 10:
                    best_params_list.pop()
            except Exception as e:
                print(f"Error with parameters {params}: {e}")
                # raise e
                continue

            if i % 50 == 0:
                self._print_intermediate_results(i, best_params_list)

        elapsed_time = time.time() - start_time
        print(f"Tuning completed in {elapsed_time:.2f} seconds")
        return best_params_list

    # def aggregate_metrics(metrics: dict) -> float:
    #     return metrics['final_score']

    def _attempt_clustering(self, embeddings, params, uuids, expected_labels, outlier_detector) -> tuple:
        hdbscan_params, outlier_params, dim_reducer_params = self._split_param_dict(params)
        reduced_embeddings = self._apply_dimensionality_reduction(embeddings, dim_reducer_params)
        predicted_labels = self._apply_clustering_and_outlier_detection(reduced_embeddings, hdbscan_params, outlier_detector, outlier_params)
        metrics = evaluate(uuids, predicted_labels.tolist(), expected_labels, reduced_embeddings)
        # print(f"Score: {metrics['final_score']}")
        return metrics, predicted_labels

    def _split_param_dict(self, params: dict) -> tuple:
        hdbscan_param_dist, outlier_param_dist, _, _, _ = self._define_param_distributions()
        hdbscan_params = {k: v for k, v in params.items() if k in hdbscan_param_dist}
        outlier_params = {k: v for k, v in params.items() if k in outlier_param_dist}
        dim_reducer_params = {k: v for k, v in params.items() if 'reducer' in k or k.startswith(params['reducer'])}
        return hdbscan_params, outlier_params, dim_reducer_params

    def _apply_dimensionality_reduction(self, embeddings: np.ndarray, params: dict) -> np.ndarray:
        if params['reducer'] == 'pca':
            pca = PCA(n_components=params['pca__n_components'])
            return pca.fit_transform(embeddings)
        elif params['reducer'] == 'umap':
            umap_reducer = umap.UMAP(n_neighbors=params['umap__n_neighbors'], min_dist=params['umap__min_dist'])
            return umap_reducer.fit_transform(embeddings)
        elif params['reducer'] == 'none':
            return embeddings
        return embeddings

    def _apply_clustering_and_outlier_detection(
        self,
        embeddings: np.ndarray,
        hdbscan_params: dict,
        outlier_detector: Optional[Callable],
        outlier_params: dict
    ) -> np.ndarray:
        clusterer = HDBSCAN(**hdbscan_params)
        clusterer.fit(embeddings)
        labels = self._assign_labels(clusterer)
        labels = self._detect_outliers_and_adjust_labels(embeddings, labels, outlier_detector, outlier_params)
        return labels

    def _detect_outliers_and_adjust_labels(self, embeddings, labels, outlier_detector, outlier_params) -> np.ndarray:
        outlier_scorer = CosineSimililarityClusterScorer()
        if outlier_detector:
            outlier_mask, item_scores = outlier_detector.detect_outliers(embeddings, labels)
        else:
            item_scores = outlier_scorer.calculate_scores(embeddings, labels)
            outlier_mask = np.full(len(labels), False)

        if outlier_detector:
            outlier_detector.threshold = outlier_params['threshold']
            outlier_detector.min_score = outlier_params['min_score']
            outlier_mask, _ = outlier_detector.detect_outliers(embeddings, labels)
            labels = np.where(outlier_mask, -1, labels)
        return labels

    def _print_intermediate_results(self, iteration: int, best_params_list: list):
        print("Top 10 Best Parameters at iteration", iteration)
        for idx, entry in enumerate(best_params_list):
            print(f"Rank {idx+1}: Score {entry['score']}, Parameters: {entry['params']}\nMetrics: {entry['metrics']},")
            print()

    def _print_final_results(self, best_params_list: list):
        print("! Final - Top 10 Best Parameters: !")
        for idx, entry in enumerate(best_params_list):
            print(f"Rank {idx+1}: Score {entry['score']}, Parameters: {entry['params']}")


from sklearn.metrics import silhouette_score
from typing import Optional
import numpy as np


from ..abstract.ClusterScorer import ClusterScorer
from ..abstract.OutlierDetector import OutlierDetector


class IQROutlierDetector(OutlierDetector):
    def __init__(
        self,
        outlier_scorer: ClusterScorer,
        threshold_range: tuple[float, float],
        step: float,
    ):
        super().__init__(outlier_scorer)
        self.threshold_range = threshold_range
        self.step = step

    def detect_outliers(
        self, embeddings: np.ndarray, labels: list[int]
    ) -> tuple[Optional[np.ndarray], dict[int, list[float]]]:
        scores = self.outlier_scorer.calculate_scores(embeddings, labels)
        best_threshold = None
        best_silhouette_score = -1
        best_outlier_mask = None

        for threshold in np.arange(
            self.threshold_range[0], self.threshold_range[1] + self.step, self.step
        ):
            outlier_mask = np.full(len(labels), False)
            for label in np.unique(labels):
                cluster_mask = labels == label
                cluster_scores = scores[label]  # Access scores by label

                q1 = np.percentile(cluster_scores, 25)
                q3 = np.percentile(cluster_scores, 75)
                iqr = q3 - q1

                lower_bound = q1 - threshold * iqr
                upper_bound = q3 + threshold * iqr
                mean_score = np.mean(cluster_scores)

                cluster_outliers = cluster_scores < lower_bound / mean_score
                outlier_mask[cluster_mask] = cluster_outliers

            silhouette_avg = silhouette_score(
                embeddings[~outlier_mask], labels[~outlier_mask]
            )
            if silhouette_avg > best_silhouette_score:
                best_silhouette_score = silhouette_avg
                best_threshold = threshold
                best_outlier_mask = outlier_mask

        print(f"Best threshold: {best_threshold:.2f}")
        print(f"Best Silhouette Score: {best_silhouette_score:.2f}")

        return best_outlier_mask, scores



import re

from .abstract.DataPreprocessor import DataPreprocessor


class TextPreprocessor(DataPreprocessor):
    def __init__(self, config: dict):
        self.config = config

    def preprocess(self, data: list[str]) -> list[str]:
        preprocessed_data = []
        for doc in data:
            doc = str(doc).lower()
            doc = doc.strip()
            doc = re.sub("</?.*?>", " <> ", doc)
            doc = re.sub(r"[^Ⴀ-ჿⴀ-ⴥᲐ-Ჿ0-9a-zA-Z.;:]", " ", doc)
            doc = re.sub(r"\s+", " ", doc)
            doc = doc.strip()
            preprocessed_data.append(doc)
        return preprocessed_data



import os
import numpy as np
from sentence_transformers import SentenceTransformer

from typing import cast, Any


from ..abstract.EmbeddingGenerator import EmbeddingGenerator


class SentenceTransformerEmbeddingGenerator(EmbeddingGenerator):
    def __init__(self, config: dict, model_name: str):
        self.config = config
        models_cache_dir = os.environ.get("MODELS_CACHE_DIR")
        if (models_cache_dir is None):
            raise Exception("MODELS_CACHE_DIR environment variable is not set")
        model_path = os.path.join(models_cache_dir, model_name)
        self.model = SentenceTransformer(model_path)

    def generate_embeddings(self, data: list[str]) -> np.ndarray:
        embeddings = self.model.encode(data)
        # FIXME: types...
        return cast(Any, embeddings)



from sklearn.metrics.pairwise import cosine_similarity
import numpy as np


from ..abstract.ClusterScorer import ClusterScorer


class CosineSimililarityClusterScorer(ClusterScorer):
    def calculate_scores(
        self, embeddings: np.ndarray, labels: np.ndarray
    ) -> dict[int, list[float]]:
        unique_labels = np.unique(labels)
        scores = {}
        for label in unique_labels:
            cluster_mask = labels == label
            cluster_embeddings = embeddings[cluster_mask]
            cluster_scores = self._calculate_cosine_similarity_scores(
                cluster_embeddings
            )
            scores[label] = cluster_scores
        return scores

    def _calculate_cosine_similarity_scores(
        self, embeddings: np.ndarray
    ) -> list[float]:
        if len(embeddings) == 1:
            return [1.0]
        similarities = cosine_similarity(embeddings)
        np.fill_diagonal(similarities, 0)
        scores = np.sum(similarities, axis=1) / (len(embeddings) - 1)
        return scores.tolist()



from typing import Optional
import numpy as np
import pandas as pd

from .abstract import ClusterScorer, OutlierDetector
from .ClusteringPipeline import ClusteringPipeline
from .embedding_generators.SentenceTransformerEmbeddingGenerator import SentenceTransformerEmbeddingGenerator

class DailyClusteringPipeline(ClusteringPipeline):
    def __init__(
        self,
        config: dict,
        model_name: str,
        uuid_col: str,
        datetime_col: str,
        outlier_scorer: ClusterScorer,
        outlier_detector: OutlierDetector,
        title_col: Optional[str] = None,
        description_col: Optional[str] = None,
        content_col: Optional[str] = None,

        verbose: bool = False,
    ):
        super().__init__(
            config=config,
            model_name=model_name,
            uuid_col=uuid_col,
            datetime_col=datetime_col,
            outlier_scorer=None,
            outlier_detector=None,
            title_col=title_col,
            description_col=description_col,
            content_col=content_col,
            verbose=verbose
        )
        self.embedding_generator = SentenceTransformerEmbeddingGenerator(config, model_name)
        self.precomputed_embeddings = None

 
    def compute_embeddings(self, data: pd.DataFrame, batch_size: int = 1000):
        # Initialize a list to store all embeddings
        all_embeddings = []

        # Get the total number of articles
        total_articles = len(data)

        # Iterate over the data in batches
        for start_idx in range(0, total_articles, batch_size):
            end_idx = min(start_idx + batch_size, total_articles)
            batch_data = data.iloc[start_idx:end_idx]

            # Generate embeddings for the current batch
            content_list = batch_data[self.title_col].fillna('').replace('\n', ' ').tolist()
            batch_embeddings = self.embedding_generator.generate_embeddings(content_list)

            # Append batch embeddings to the list
            all_embeddings.append(batch_embeddings)

            if self.verbose:
                print(f"Processed batch {start_idx // batch_size + 1}/{(total_articles - 1) // batch_size + 1}")

        # Concatenate all batch embeddings
        self.precomputed_embeddings = pd.DataFrame(np.vstack(all_embeddings), index=data[self.uuid_col])

        if self.verbose:
            print("Embeddings computed for all articles.")


    def run_daily_clustering(self, data: pd.DataFrame):
        # Ensure embeddings are precomputed
        if self.precomputed_embeddings is None:
            raise ValueError("Embeddings are not precomputed. Call compute_embeddings first.")

        # Group data by date
        data['date'] = data[self.datetime_col].dt.date
        daily_groups = data.groupby('date')

        for date, group in daily_groups:
            print(f"Processing date: {date}")
            # Retrieve embeddings for the current group
            group_embeddings = self.precomputed_embeddings.loc[group[self.uuid_col]]
            # Run clustering directly with precomputed embeddings
            labels = self.run(group, embeddings=group_embeddings.values, process_embeddings=True)
            # Store results
            self.results[date] = {
                'data': group,
                'labels': labels
            }

    def get_results(self):
        return self.results

# Usage
# Assuming you have a DataFrame `df` with the required columns